# TP Hachage


# Partie 1 - Réaliser une fonction de hachage minimale parfaite 

On travaillera sur `tp_2_miso_mphf.py`.

Tout d'abord, voyons ensemble l'idée de la construction de la MPHF.
Puis suivez les commentaires dans le code à compléter.
Vous devez 
- finaliser la fonction pour construire la MPHF (`construction_mphf`), 
- écrire la fonction pour obtenir le hash d'un élément par la MPHF (`get_hash_mphf`), 
- écrire la fonction pour créer une table de hachage avec cette mphf (`create_hash_table`)

Ecrivez vos réponses et commentaires dans ce document.

### Explication de tableaux.append(tableau_principal)

Cette ligne permet d'ajouter le tableau principal (actuel) à la liste des tableaux. 
On fait cela car on va réduire la taille de la table de hachage à plusieurs reprise, nous allons donc stocker dans cette liste les table au fur et à mesure.

### Explication de set_kmer_courant = collision.copy()

Après un niveau il se peut qu'il y ait des k-mers en collision ceux-ci sont enregistré dans 'collision'.
Ensuite set_kmer_courant est mis à jour pour qu'il contienne les k-mers qui ont subi une collison. 
L'objectif est d'essayer d'insérer ces k-mer en collision pour les intégrer dans le prochain niveau de la table de hachage.

### Explication de collision = set()

Cette ligne permet de vider collision. En effet après avoir copié les valeurs de collision dans set_kmer_courant, 
on va réinitialiser collision pour le prochain niveau de la table de hashage. 
On recommence avec une liste collision qui contient les éléments à traiter.

### Explication de grand_tableau.extend(tableau)

Nous avons des tableaux crée à chaque niveau pour contenir les résultats intermédiaire du hachage.
Cette ligne permet d'ajouter tout les éléments qui sont stockés dans 'tableau' dans seul grand_tableau. 
On va ainsi pouvoir regrouper toute les tables intermédiaires dans un seul grand tableau et simplifier l'accès aux indices des k-mers.

### Explication de:
for kmer in set_kmer_courant:
max_rang += 1
h = abs(hash(kmer))
mphf.append([h, max_rang])

Ce bloc permet de donner un rang aux k-mers, celui sera donc unique pour les cas où des collisions persisterais à la fin.
On parcours tous les kmers et on leurs attribut un rang +1 par rapport au précédent.
On recupère le hash du k-mer puis on l'ajoute avec le rang à la MPHF

### Explication de pourquoi ici on ne mesure pas juste la taille en mémoire du tableau ? 
table_size.append(sys.getsizeof(tableau) + sys.getsizeof(mphf))

Cette ligne permet de calculer la taille de tableau qui contient les k-mers placé à la position de leurs hash respectif et est
additioné à la taille de la mphf qui contient la fonction pour bien placer ces k-mers.
Il nous faut pour avoir une bonne estimation de l'utilisation de la mémoire prendre le tableau et la mphf.
Nous pourrons ensuite comparer avec la taille des dictionnaire python.

Puis décommenter `compare_taille` à la fin et expliquer les résultats.

### Explication du graphique mphf.png

Ce graphique permet d'observer l'évolution de la taille mémoire utilisée par une table MPHF (en bleu) et un dictionnaire python en orange pour un nombre 'n' de k-mer.
On observe que la courbe MPHF augmente progressivement de manière linéaire tandis que la courbe 'dict' augmente en escalier.
La courbe augmente par palier brutalement vers 4100 k-mers.
Cela pourrait s'expliquer par le fait que les dictionnaire python vont doubler la taille de la mémoire pour éviter les phénomènes de collision.
Nous pouvons conclure que la MPHF consomme beaucoup moins de mémoire que le dictionnaire surtout pour les grandes valeurs.


Bonus : faites varier `nb_niveaux` et `gamma`, voyez quelle influence ils peuvent avoir.

# Partie 2 - Analyse de performance de dictionnaires en Python

On travaillera sur `tp_2_miso_dict.py`

A la fin de cette partie, en lançant `python tp_2_miso_dict.py` on doit obtenir 4 graphiques.

## Introduction

Nous allons analyser la performance de dictionnaires `dict` en Python en fonction de différents facteurs : que le facteur de charge (loading factor), 
le temps d'insertion et la taille de la mémoire occupée. Vous pourrez utiliser matplotlib et numpy pour visualiser les résultats.

##  Étude du facteur de charge

Fonction `experiment_load_factor`:

0. La fonction `experiment_load_factor` doit être définie pour prendre en entrée une liste de facteurs de charge (`load_factors`) et renvoyer les temps 
- d'insertion de clefs, 
- les nombres de réallocations de mémoire (quand le dictionnaire est ré-écrit en mémoire pour agrandir sa taille) 
- et les tailles de mémoire occupées par le dictionnaire pour chaque facteur de charge. 

1. Initialisation des listes pour stocker les résultats.
Initialiser dans la fonction `experiment_load_factor` des listes `insertion_times`, `num_resizes` et `sizes` seront utilisées pour stocker les trois résultats décrits au dessus.

2. La fonction `experiment_load_factor` doit boucler sur chaque facteur de charge de la liste en entrée `load_factors`. Créer un dictionnaire d'abord vide à chaque fois.

3. Initialiser les variables `num_elements`, `start_time`, `num_resize` et `last_size` pour respectivement mesurer  le nombre d'éléments, le temps d'insertion et le nombre de réallocations de mémoire.

4. Insérer les éléments dans le dictionnaire et mesurer le temps d'insertion pour chaque élément, vérifier le nombre de réallocations mémoire (utiliser par exemple `time.time()` et `sys.getsizeof()` pour mesurer le temps et la taille du dictionnaire avant et après chaque insertion).

5. Pour un facteur de charge donné, stocker les résultats dans les listes `insertion_times`, `num_resizes` et `sizes`.

## Deuxième étude

6. A quoi sert la fonction `experiment_longest` ?

La fonction experiment_longest parcourt, par une itération, les nombres de 0 à 999 et stocke chaque valeur dans un dictionnaire 
avec une clé correspondant au nombre sous forme de chaîne de caractères.
À chaque insertion d'un élément dans le dictionnaire, le temps d'insertion est mesuré et ajouté dans un histogramme.

## Visualisation des résultats

7. Créez quatre graphiques au format png ou pdf : 

- Un graphique du temps d'insertion en fonction du facteur de charge (obtenu question 5)
- Un graphique du nombre de réallocations de mémoire en fonction du facteur de charge (obtenu question 5)
- Un graphique de la taille de mémoire occupée en fonction du nombre d'éléments (obtenu question 5)
- Un histogramme des fréquences des temps d'insertions discrétisés (code fourni, remplacer la liste vide par la bonne entrée)

10. Commentez vos résultats.

7A : Ce graphique représente le temps d'insertion des éléments dans le dictionnaire en fonction du facteur de charge. 
Nous avons calculé la moyenne du temps d'insertion pour chaque facteur de charge et l'avons représentée sur le graphique.
Nous nous attendions à ce que le temps d'insertion soit plus faible pour les petits facteurs de charge, mais nous avons observé l'inverse, sans pouvoir l'expliquer. 
Nous avons cherché un éventuel biais en utilisant la médiane au lieu de la moyenne, mais les résultats sont restés identiques.
Nous pouvons cependant noter qu'un dictionnaire possède une taille initiale définie. 
Lorsqu'un élément est ajouté, il faut parcourir le dictionnaire et trouver un emplacement libre, ce qui explique le temps d'insertion observé, qui est censé être plus grand avec le facteur de charge.

7B : Ce graphique illustre le nombre de réallocations de mémoire en fonction du facteur de charge. 
On remarque que les petits facteurs de charge nécessitent peu d'espace, et donc peu de réallocations.
En revanche, pour les plus grands facteurs de charge, la mémoire requise augmente, entraînant davantage de réallocations. 
On observe également des plateaux dans le graphique. En effet, la taille du dictionnaire double à chaque réallocation, 
laissant un espace libre suffisant pour insérer de nouveaux éléments sans nécessiter immédiatement une nouvelle réallocation. 
Ainsi, on insère jusqu'à saturation, puis une nouvelle réallocation est déclenchée.

7C : Ce graphique représente la taille de la mémoire occupée en fonction du nombre d'éléments insérés.
On observe que pour les petites valeurs, le nombre de réallocations de mémoire est faible. 
À mesure que le facteur de charge augmente, la taille des réallocations croît également. 
Cependant, on note des écarts importants à certains moments. Cela s'explique par le mécanisme des réallocations : la taille du dictionnaire double lorsque l'espace est insuffisant. 
Par conséquent, pour certaines valeurs du facteur de charge, il reste suffisamment d’espace pour insérer plusieurs éléments sans nécessiter de nouvelle réallocation.

Histogramme.png : Ce dernier graphique montre la distribution des temps d'insertion sous forme d'histogramme.
On remarque que la plupart des valeurs sont insérées très rapidement dans le dictionnaire, tandis qu'un faible nombre d'insertions prennent plus de temps. 
Cela s’explique par le fait que le dictionnaire doit parfois parcourir son espace de stockage pour trouver une position libre avant d'insérer la nouvelle clé et sa valeur. 
